## This code calculates via simulation weights for airborne laser scanning
## pulses that are subsequently used in estimating the trajectory of the laser
## instrument.
##
## The user must specify values for the following simulation parameters:
## * Range of laser return coordinate imprecision induced by rounding. If
##   coordinates are rounded to two digits of precision, the range should be
##   [-0.005, 0.005]. If to three digits of precision [-0.0005, 0.0005].
## * Number of simulation repetitions. 10000 would suffice for most cases
## * Vector of above-target platform height values. Appropriate values can
##   often be determined from data vendor delivery reports. 
## * Vector of values representing the distance, along the pulse trajectory,
##   between first and last reports. Note that a value of zero would be
##   because for technical reasons returns generated by the same pulse are
##   separated by a minimum distance which varies by laser instrument but 
##   usually ranges between 1.5 and 2 meters. 
## * Vector of pulse view angle values. Those should be greater than 0.0 and no
##   larger than 45.0 degrees (No negative values). Pulse angles should not be
##   confused with scan angles recorded per return in raw laser files. Upper
##   range of pulse angles could be 10 degrees higher than the 95th percentile
##   of recorded scan angles.
##
## For computational efficiency selected functions are embedded in cmp_trj.dll
## The file must be present in the working directory and it is a 64-bit version.
## Please contact the authors if you need a 32-bit version. The simulation is
## executed in parallel, using twice as many threads as the physical cores of
## the computer.
##
## Calculated weights are printed on the console. They can then be appended to
## cmp_trj.R program which computes the laser scanner's trajectory for a set of
## scan swaths. Note that the output weights are negative. cmp_trj.R needs 
## their absolute values.
##
## More details are available in our paper titled "Reconstructing aircraft
## trajectories from multi-return airborne laser scanning data"
##
## July 2019

## ------------------ USER INPUT ------------------

error.rng <- c(-0.005, 0.005)       ## range of return coordinate error induced
                                    ## by rounding. 0.05 for 1-digit rounding,
                                    ## 0.005 for 2-digit, etc.
reps      <- 100000                 ## Number of repetitions
HT        <- seq(500, 2000, by=100) ## above-ground platform heights
D         <- seq(2, 40, by=2)       ## distances b/w 1st and last return
theta     <- seq(1, 30, by=1)       ## view angles (degrees) [> 0.0 & << 90.0]

## --------------- END OF USER INPUT --------------

calc.error.vec.C <- function(n, HT, D, theta, dist.vec) {
  Cout <- .C( "assessErrorUniform", 
              as.integer(n), as.double(runif(12*n, error.rng[1], error.rng[2])),
              as.double(HT), as.double(D), as.double(theta),
              as.double(dist.vec) )
  return( Cout[[6]])
}

## ------------------------------------------------

suppressWarnings( suppressMessages( c(library(parallel), library(doSNOW)) ) )

require(doSNOW)
require(parallel)

## cmp_trj.dll required in working directory

cat(paste("\nCalculating pulse weights ... \n\n"))

combs <-  expand.grid(HT=HT, D=D, THETA=theta)

cl <- makeCluster(detectCores(logical = TRUE))
registerDoSNOW(cl)

pb <- txtProgressBar(max = nrow(combs), style = 3, char=".")
progress <- function(n) setTxtProgressBar(pb, n)
opts <- list(progress = progress)

error.mat <- foreach(j=1:nrow(combs), .combine="rbind", .options.snow = opts) %dopar% {
   dyn.load("cmp_trj.dll")
   HT    <- combs$HT[j]
   D     <- combs$D[j]
   theta <- combs$THETA[j]
   ## Call to C function uses theta, not theta.rad
   error.vec <- calc.error.vec.C(reps, HT, D, theta, rep(-9999.0, reps))
   c(HT, D, theta, round(mean(error.vec),3), round(sd(error.vec),3))
}

close(pb)
stopCluster(cl)

error.df <- as.data.frame(error.mat)
names(error.df) <- c("HT", "D", "THETA", "MEAN", "SD")

summary(error.df)

the.lm <- lm(log(MEAN) ~ log(HT) + log(D) + log(THETA), data=error.df)
summary(the.lm)

cat("\nWeight of distance b/w 1st and last return:", round(the.lm$coef[3],4), "\n")
cat("Weight of view angle:                      ", round(the.lm$coef[4],4), "\n\n")
