## This code estimates the trajectory of a laser instrument onboard an airborne platform
## for a set of las or laz files of point clouds representing individual scan swaths. The 
## user must specify values for all input parameters listed below.
## Optimal values for SEGLENFactor and AngleFactor can be optained via simulation enabled
## by calc_pulse_weights.R
##
## The output includes
##  * X, Y, Z, and GPSTime values of the trajectory calculated for each individual scan
##    swath in user-specified frequency (if requested), and
##  * a .csv (text) file with cubic spline fit metrics
##
## A detailed description of the process is available in our paper titled "Reconstructing
## aircraft trajectories from multi-return airborne laser scanning data"
##
## July 2019

## ------------------ USER INPUT ------------------

las.list        <- "list.asc" ## dir /b/s *.laz > list.asc at DOS prompt creates list with path
                              ##   swath files assumed named swath_xxxx.las or swath_xxxx.laz
multi.pulse     <- FALSE      ## Logical. TRUE only for systems with multiple pulses in the air
SEGLENFactor    <- 1.0059     ## Weighting factor for the distance b/w 1st and last pulse returns
AngleFactor     <- 0.8824     ## Weighting factor for view angle of mother pulse of a return
deltaT          <- 0.05       ## TimeBlock duration (in seconds)
CPAInterval     <- 4          ## Time between CPAs considered in fitting a cubic spline (in seconds)
dist.vec        <- c( 500, 300, 200, 150, 100, 75, 50, 25, 0 ) ## in meters. Used to remove spline fit outliers
trim.deltaT     <- 3          ## end-of-swath return trimming interval (in seconds). Set to 0 for no trimming
unit            <- "FT"       ## either "FT" or "M"
exportFrequency <- 100        ## locations of assessed trajectory (Hz)
exportType      <- "DBF"      ## options are "NONE", TEXT", or "DBF"
outputName      <- "fit.csv"  ## Whete to save cubic spline fit results. Must have .csv extension

## --------------- END OF USER INPUT --------------

suppressWarnings( suppressMessages( c(library(data.table), library(plyr),
                                      library(lidR), library(foreign),
                                      library(parallel), library(doSNOW)) ) )
library(data.table)
library(plyr)
library(lidR)
library(foreign)
library(parallel)
library(doSNOW)

source("cmp_trj_functions.R")
dyn.load("cmp_trj.dll")

SMALL.FLT <- 1e-8

if(!file.exists(las.list)) stop(paste("Specified file", las.list, "does NOT exist\n\n"))
file.list  <- read.table(las.list)
n.files    <- nrow(file.list)

cat(paste("Calculating platform trajectories for", n.files, "files ... \n\n"))

cl <- makeCluster(2, type="SOCK") ## increase number of threads if reading from SSD or striped RAID
registerDoSNOW(cl)
export.list <- list("readLAS", "pulse2trj", "fitAndEvalSpline", "count", "exportSpline", "write.dbf")
clusterExport(cl=cl, export.list)

pb       <- txtProgressBar(max=n.files, style=3, char=".")
progress <- function(n) setTxtProgressBar(pb, n)
opts     <- list(progress = progress)

t1 <- Sys.time()

output.list <- foreach(i=1:n.files, .packages=c("data.table", "plyr"), .options.snow = opts) %dopar% {
  dyn.load("cmp_trj.dll")
  swath.file <- as.character(file.list[i,1])
  if( !file.exists(swath.file) ) {
    stop(paste("File", swath.file, "does NOT exist\n\n"))
  }

  swath.name <- substr(swath.file, 1, nchar(swath.file)-4) ## assumes .las or .laz extension
  flightID <- substr(swath.name, max(gregexpr("_", swath.name)[[1]])+1, nchar(swath.name))
  
  if( multi.pulse == F ) { ## order by gpstime and ReturnNumber
    las <- readLAS(swath.file, select="tra") ## t=gpstime, r=ReturnNumber, a=ScanAngle
    las@data <- setorder(las@data, gpstime, ReturnNumber) ## set record order
    las <- las@data
    ## .C returns a list of a number, a 2nd number, a gpstime vector, a 3rd number,
    ##  and a flag vector. I am picking the fifth
    ## las must be sorted by gpstime and ReturnNumber
    removeReturnsFlag <- .C( "keepFirstLastPulseReturns", as.integer(nrow(las)), as.integer(1),
                             as.double(las$gpstime), as.integer(0),
                             as.integer(rep(-1, nrow(las))) )[[5]]
  } else { ## order by gpstime, UserData, and ReturnNumber
    las <- readLAS(swath.file, select="trauRGB") ## t=gpstime, r=ReturnNumber, a=ScanAngle, u=UserData
    las@data <- setorder(las@data, gpstime, UserData, ReturnNumber) ## set record order
    las <- las@data
    ## .C returns a list of a number, a 2nd number, a gpstime vector, a userData vector,
    ##  and a flag vector. I am picking the fifth
    ## las must be sorted by gpstime, UserData and ReturnNumber
    removeReturnsFlag <- .C( "keepFirstLastPulseReturns", as.integer(nrow(las)), as.integer(2),
                             as.double(las$gpstime), as.integer(las$UserData),
                             as.integer(rep(-1, nrow(las))) )[[5]]
  }
  las <- las[removeReturnsFlag == 1,]

  ## rearrange las into data table with one record per pulse, and add items SCANANGLE and SEGLEN
  v <- seq( 2, nrow(las), by=2 )
  pulse.dt <- as.data.table(las[v, c("X", "Y", "Z")])
  names(pulse.dt) <- c("XFIRST", "YFIRST", "ZFIRST")
  pulse.dt[, c("XLAST", "YLAST", "ZLAST", "T", "SCANANGLE"):= las[v-1, c("X", "Y", "Z", "gpstime", "ScanAngleRank")]]
  pulse.dt[, SEGLEN:= sqrt( (XFIRST - XLAST)^2 + (YFIRST - YLAST)^2 + (ZFIRST - ZLAST)^2 ) ]
  rm(v, las); gc()

  trj.df    <- pulse2trj(pulse.dt, deltaT)  
  tSpan     <- max(trj.df$T) - min(trj.df$T)
  knots.vec <- sort(unique(as.integer(round(tSpan/CPAInterval))), decreasing=T)
  knots.vec <- knots.vec[knots.vec > 3 & knots.vec < nrow(trj.df)]
  n.knots   <- length(knots.vec)
  
  minT <- min(trj.df$T)
  maxT <- max(trj.df$T)
  
  if( n.knots > 0 & nrow(trj.df) > 4 ) {
    fit.list <- fitAndEvalSpline( flightID, trj.df, knots.vec, dist.vec, unit )
    TRRMSE <- round( fit.list[[2]], 4 )
    RMSE   <- round( fit.list[[3]], 4 )
    if( toupper(exportType) != "NONE" ) {
      exportSpline(fit.list[[4]], fit.list[[5]], fit.list[[6]], min(trj.df$T), max(trj.df$T), exportFrequency, exportType, flightID)
    }
    trj.df <- fit.list[[1]]
    c(flightID, deltaT, CPAInterval, TRRMSE, RMSE)
  } else {
    c(flightID, deltaT, CPAInterval, NA, NA)
  }
}

close(pb)
stopCluster(cl)

output.df <- data.frame(do.call(rbind, output.list))
names(output.df) <- c("FLIGHTID", "DELTAT", "CPAINTERVAL", "TRRMSE", "RMSE")

dyn.unload("cmp_trj.dll")

write.csv(output.df, file=outputName, row.names=F)

t2 <- round(Sys.time() - t1,2)
cat("\nElapsed time:", t2, attributes(t2)$units, "\n")
